"""
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import hashlib
import json
import logging
from datetime import datetime
from typing import Any, Optional
from urllib import parse

from urllib3 import PoolManager
from urllib3.exceptions import HTTPError

from sleet.constants import API_VALID_REGIONS, DEFAULT_CLIENT_HEADERS
from sleet.exceptions import APIStatusError, RegionError, SleetError
from sleet.oauth import OAuth

logger = logging.getLogger(__name__)


class BaseAPI:
    hashing_enabled: bool = False

    def __init__(self, *, region: str, client_id: str, client_secret: str):
        self.api_region = region.lower()
        if self.api_region not in API_VALID_REGIONS:
            raise RegionError(self.api_region)

        self.api_url = f"https://{self.api_region}.api.blizzard.com".lower()
        self.client: PoolManager = PoolManager(headers=DEFAULT_CLIENT_HEADERS)
        self.oauth = OAuth(self.api_region, client_id, client_secret)

    def request(
            self,
            url: str,
            parameters: Optional[dict[str, Any]],
            headers: Optional[dict[str, str]]
    ) -> dict[str, Any]:
        """
        TODO: Fill in.

        :param url:
        :param parameters:
        :param headers:
        :return:
        """
        resource_url = f"{self.api_url}{url}"
        if parameters:
            resource_url = f"{resource_url}?{parse.urlencode(parameters)}"

        headers = headers or {}
        headers = headers | self.client.headers | {"Authorization": f"Bearer {self.oauth.token}",
                                                   "X-Sleet-Timestamp": f"{int(datetime.utcnow().timestamp())}"}

        logger.debug(f"Requesting {resource_url}. Headers: {headers}. Parameters: {parameters}.")

        try:
            response = self.client.request("GET", resource_url, headers=headers)

        except HTTPError as ex:
            raise SleetError(f"Error querying {resource_url}.") from ex

        if response.status != 200:
            raise APIStatusError(response.status)

        return self.response_unpack(resource_url, headers, response)

    def response_unpack(self, url: str, headers: dict[str, str], response: Any) -> dict[str, Any]:
        time_taken = datetime.utcnow().timestamp() - int(headers["X-Sleet-Timestamp"])

        metadata = {
            "url": url,
            "headers": {
                "request": headers,
                "response": dict(response.headers)
            },
            "sha256": None,
            "elapsed": round(time_taken, 3)
        }

        if self.hashing_enabled:
            m = hashlib.sha256()
            m.update(response.data)
            metadata["sha256"] = m.hexdigest()

        response = response.data.decode()

        try:
            response = json.loads(response)

        except json.JSONDecodeError as ex:
            logger.error(f"Failed to obtain new token. Exception: {ex}.")
            raise SleetError(f"Failed to obtain new token. Exception: {ex}.")

        response["@metadata"] = metadata

        return response

    @staticmethod
    def create_modified_since_header(modified_since: Optional[datetime]) -> dict[str, str]:
        if not isinstance(modified_since, datetime):
            raise TypeError(f"Supplied modified_since '{modified_since}' is not a valid datetime.")

        return {"If-Modified-Since": modified_since.strftime("%a, %d %b %Y %H:%M:%S GMT")}
