"""
This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
If a copy of the MPL was not distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""

import inspect
from datetime import datetime
from typing import Any, Iterable, Optional

from sleet.base import BaseAPI
from sleet.constants import Namespace, WarcraftExpansion
from sleet.exceptions import SleetError, WarcraftExpansionError


# noinspection GrazieInspection
class WarcraftAPI(BaseAPI):

    def __init__(self, region: str, *, client_id: str, client_secret: str, expansion: WarcraftExpansion):
        super().__init__(region=region, client_id=client_id, client_secret=client_secret)
        self.expansion = expansion

    def prepare(
            self,
            *,
            expansions: Iterable[WarcraftExpansion],
            namespace: Namespace,
            url: str,
            headers: Optional[dict[str, str]] = None,
            parameters: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """
        TODO: Fill in.

        :param expansions:
        :param namespace:
        :param url:
        :param headers:
        :param parameters:
        :return:
        """
        if self.expansion not in expansions:
            raise WarcraftExpansionError(inspect.stack()[1][3], self.expansion, expansions)

        if self.expansion == WarcraftExpansion.CLASSIC:
            ns = f"{namespace.value}-classic1x"
        elif self.expansion == WarcraftExpansion.TBCC:
            ns = f"{namespace.value}-classic"
        else:
            ns = namespace.value

        headers = (headers or {}) | {"Battlenet-Namespace": f"{ns}-{self.api_region}"}

        return self.request(url, parameters, headers)

    # Achievement APIs
    def achievement_categories_index(self) -> dict[str, Any]:
        """
        Returns an index of achievement categories.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/achievement-category/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def achievement_category(self, achievement_category_id: int) -> dict[str, Any]:
        """
        Returns an achievement category by ID.

        :param achievement_category_id: The ID of the achievement category.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/achievement-category/{achievement_category_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def achievement_index(self) -> dict[str, Any]:
        """
        Returns an index of achievements.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/achievement/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def achievement(self, achievement_id: int) -> dict[str, Any]:
        """
        Returns an achievement by ID.

        :param achievement_id: The ID of the achievement.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/achievement/{achievement_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def achievement_media(self, achievement_id: int) -> dict[str, Any]:
        """
        Returns media for an achievement by ID.

        :param achievement_id: The ID of the achievement.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/achievement/{achievement_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Auction APIs
    def auctions(self, connected_realm_id: int, modified_since: Optional[datetime] = None) -> dict[str, Any]:
        """
        Returns all active auctions for a connected realm.

        Depending on the number of active auctions on the specified connected realm,
        the response from this endpoint may be large, sometimes exceeding 10 MB.

        :param connected_realm_id: The ID of the connected realm.
        :param modified_since: Only return data if it's newer than the datetime specified
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/connected-realm/{connected_realm_id}/auctions"

        headers = {}
        if modified_since:
            headers = self.create_modified_since_header(modified_since)

        return self.prepare(expansions=expansions, namespace=namespace, url=url, headers=headers)

    def auction_index(self, connected_realm_id: int) -> dict[str, Any]:
        """
        Returns the auction index's for a connected realm.

        :param connected_realm_id: The ID of the connected realm.
        """
        expansions = WarcraftExpansion.TBCC,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/connected-realm/{connected_realm_id}/auctions/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def auction_house(
            self,
            connected_realm_id: int,
            auction_house: str,
            modified_since: Optional[datetime] = None
    ) -> dict[str, Any]:
        """
        Returns all active auctions for a connected realm at a specific auction house.

        Depending on the number of active auctions on the specified connected realm,
        the response from this endpoint may be rather large, sometimes exceeding 10 MB.

        :param auction_house: ALLIANCE, HORDE, or NEUTRAL.
        :param connected_realm_id: The ID of the connected realm.
        :param modified_since: Only return data if it's newer than the datetime specified
        """
        expansions = WarcraftExpansion.TBCC,
        namespace = Namespace.DYNAMIC

        auction_house = auction_house.upper()
        if auction_house == "ALLIANCE":
            url = f"/data/wow/connected-realm/{connected_realm_id}/auctions/2"
        elif auction_house == "HORDE":
            url = f"/data/wow/connected-realm/{connected_realm_id}/auctions/6"
        elif auction_house in ("NEUTRAL", "BLACKWATER"):
            url = f"/data/wow/connected-realm/{connected_realm_id}/auctions/7"
        else:
            raise SleetError(f"{auction_house} is not a valid Auction House.")

        headers = {}
        if modified_since:
            headers = self.create_modified_since_header(modified_since)

        return self.prepare(expansions=expansions, namespace=namespace, url=url, headers=headers)

    # Azerite Essence APIs
    def azerite_essences_index(self) -> dict[str, Any]:
        """
        Returns an index of azerite essences.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/azerite-essence/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def azerite_essence(self, azerite_essence_id: int) -> dict[str, Any]:
        """
        Returns an azerite essence by ID.

        :param azerite_essence_id: The ID of the Azerite Essence
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/azerite-essence/{azerite_essence_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def azerite_essence_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of azerite essences.
        """
        raise NotImplementedError()

    def azerite_essence_media(self, azerite_essence_id: int) -> dict[str, Any]:
        """
        Returns media for an azerite essence by ID.

        :param azerite_essence_id: The ID of the Azerite Essence
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/azerite-essence/{azerite_essence_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Connected Realm APIs
    def connected_realm_index(self) -> dict[str, Any]:
        """
        Returns an index of connected realms.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = "/data/wow/connected-realm/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def connected_realm(self, connected_realm_id: int) -> dict[str, Any]:
        """
        Returns a connected realm by ID.

        :param connected_realm_id: The ID of the connected realm.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/connected-realm/{connected_realm_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def connected_realm_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of connected realms
        """
        raise NotImplementedError()

    # Covenant APIs
    def covenant_index(self) -> dict[str, Any]:
        """
        Returns an index of covenants.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/covenant/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def covenant(self, covenant_id: int) -> dict[str, Any]:
        """
        Returns a covenant by ID.

        :param covenant_id: The ID of the covenant.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/covenant/{covenant_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def covenant_media(self, covenant_id: int) -> dict[str, Any]:
        """
        Returns media for a covenant by ID.

        :param covenant_id: The ID of the covenant.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/covenant/{covenant_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def soulbind_index(self) -> dict[str, Any]:
        """
        Returns an index of soulbinds.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/covenant/soulbind/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def soulbind(self, soulbind_id: int) -> dict[str, Any]:
        """
        Returns a soulbind by ID.

        :param soulbind_id: The ID of the soulbind.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/covenant/soulbind/{soulbind_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def conduit_index(self) -> dict[str, Any]:
        """
        Returns an index of conduits.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/covenant/conduit/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def conduit(self, conduit_id: int) -> dict[str, Any]:
        """
        Returns a conduit by ID.

        :param conduit_id: The ID of the conduit.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/covenant/conduit/{conduit_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Creature APIs
    def creature_families_index(self) -> dict[str, Any]:
        """
        Returns an index of creature families.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/creature-family/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature_family(self, creature_family_id: int) -> dict[str, Any]:
        """
        Returns a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/creature-family/{creature_family_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature_types_index(self) -> dict[str, Any]:
        """
        Returns an index of creature types.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/creature-type/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature_type(self, creature_type_id: int) -> dict[str, Any]:
        """
        Returns a creature type by ID.

        :param creature_type_id: The ID of the creature type.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/creature-type/{creature_type_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature(self, creature_id: int) -> dict[str, Any]:
        """
        Returns a creature by ID.

        :param creature_id: The ID of the creature.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/creature/{creature_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of creatures.
        """
        raise NotImplementedError()

    def creature_display_media(self, creature_display_id: int) -> dict[str, Any]:
        """
        Returns media for a creature display by ID.

        :param creature_display_id: The ID of the creature display.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/creature-display/{creature_display_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def creature_family_media(self, creature_family_id: int) -> dict[str, Any]:
        """
        Returns media for a creature family by ID.

        :param creature_family_id: The ID of the creature family.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/creature-family/{creature_family_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Guild Crest APIs
    def guild_crest_components_index(self) -> dict[str, Any]:
        """
        Returns an index of guild crest media.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/guild-crest/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def guild_crest_border_media(self, border_id: int) -> dict[str, Any]:
        """
        Returns media for a guild crest border by ID.

        :param border_id: The ID of the guild crest border.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/guild-crest/border/{border_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def guild_crest_emblem_media(self, emblem_id: int) -> dict[str, Any]:
        """
        Returns media for a guild crest emblem by ID.

        :param emblem_id: The ID of the guild crest emblem.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/guild-crest/emblem/{emblem_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Item APIs
    def item_classes_index(self) -> dict[str, Any]:
        """
        Returns an index of item classes.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/item-class/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_class(self, item_class_id: int) -> dict[str, Any]:
        """
        Returns an item class by ID.

        :param item_class_id: The ID of the item class.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/item-class/{item_class_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_sets_index(self) -> dict[str, Any]:
        """
        Returns an index of item sets.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/item-set/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_set(self, item_set_id: int) -> dict[str, Any]:
        """
        Returns an item set by ID.

        :param item_set_id: The ID of the item set.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/item-set/{item_set_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_subclass(self, item_class_id: int, item_subclass_id: int) -> dict[str, Any]:
        """
        Returns an item subclass by ID.

        :param item_class_id: The ID of the item class.
        :param item_subclass_id: The ID of the item subclass.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/item-class/{item_class_id}/item-subclass/{item_subclass_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item(self, item_id: int) -> dict[str, Any]:
        """
        Returns an item by ID.

        :param item_id: The ID of the item.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/item/{item_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_media(self, item_id: int) -> dict[str, Any]:
        """
        Returns media for an item by ID.

        :param item_id: The ID of the item.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/item/{item_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def item_search(self, **kwargs: Any) -> dict[str, Any]:
        """
        Performs a search of items.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/search/item"

        return self.prepare(expansions=expansions, namespace=namespace, url=url, parameters=kwargs)

    # Journal APIs
    def journal_expansions_index(self) -> dict[str, Any]:
        """
        Returns an index of journal expansions.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/journal-expansion/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_expansion(self, journal_expansion_id: int) -> dict[str, Any]:
        """
        Returns a journal expansion by ID.

        :param journal_expansion_id: The ID of the journal expansion.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/journal-expansion/{journal_expansion_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_encounters_index(self) -> dict[str, Any]:
        """
        Returns an index of journal encounters.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/journal-encounter/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_encounter(self, journal_encounter_id: int) -> dict[str, Any]:
        """
        Returns a journal encounter by ID.

        :param journal_encounter_id: The ID of the journal encounter.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/journal-encounter/{journal_encounter_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_encounter_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of journal encounters.
        """
        raise NotImplementedError()

    def journal_instances_index(self) -> dict[str, Any]:
        """
        Returns an index of journal instances.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/journal-instance/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_instance(self, journal_instance_id: int) -> dict[str, Any]:
        """
        Returns a journal instance.

        :param journal_instance_id: The ID of the journal instance.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/journal-instance/{journal_instance_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def journal_instance_media(self, journal_instance_id: int) -> dict[str, Any]:
        """
        Returns media for a journal instance by ID.

        :param journal_instance_id: The ID of the journal instance.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/journal-instance/{journal_instance_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Media Search APIs
    def media_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of all types of media documents.
        """
        raise NotImplementedError()

    # Modified Crafting APIs
    def modified_crafting_index(self) -> dict[str, Any]:
        """
        Returns the parent index for Modified Crafting.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/modified-crafting/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def modified_crafting_category_index(self) -> dict[str, Any]:
        """
        Returns the index of Modified Crafting categories.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/modified-crafting/category/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def modified_crafting_category(self, category_id: int) -> dict[str, Any]:
        """
        Returns a Modified Crafting category by ID.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/modified-crafting/category/{category_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def modified_crafting_reagent_slot_type_index(self) -> dict[str, Any]:
        """
        Returns the index of Modified Crafting reagent slot types.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/modified-crafting/reagent-slot-type/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def modified_crafting_reagent_slot_type(self, slot_type_id: int) -> dict[str, Any]:
        """
        Returns a Modified Crafting reagent slot type by ID.

        :param slot_type_id: The ID of the Modified Crafting reagent slot type.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/modified-crafting/reagent-slot-type/{slot_type_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Mount APIs
    def mounts_index(self) -> dict[str, Any]:
        """
        Returns an index of mounts.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/mount/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mount(self, mount_id: int) -> dict[str, Any]:
        """
        Returns a mount by ID.

        :param mount_id: The ID of the mount.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/mount/{mount_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mount_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of mounts.
        """
        raise NotImplementedError()

    # Mythic Keystone Affix APIs
    def mythic_keystone_affixes_index(self) -> dict[str, Any]:
        """
        Returns an index of mythic keystone affixes.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/keystone-affix/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_affix(self, keystone_affix_id: int) -> dict[str, Any]:
        """
        Returns a mythic keystone affix by ID.

        :param keystone_affix_id: The ID of the mythic keystone affix.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/keystone-affix/{keystone_affix_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_affix_media(self, keystone_affix_id: int) -> dict[str, Any]:
        """
        Returns media for a mythic keystone affix by ID.

        :param keystone_affix_id: The ID of the mythic keystone affix.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/keystone-affix/{keystone_affix_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Mythic Keystone Dungeon API
    def mythic_keystone_dungeon_index(self) -> dict[str, Any]:
        """
        Returns an index of Mythic Keystone dungeons.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = "/data/wow/mythic-keystone/dungeon/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_dungeon(self, dungeon_id: int) -> dict[str, Any]:
        """
        Returns a Mythic Keystone dungeon by ID.

        :param dungeon_id: The ID of the dungeon.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/mythic-keystone/dungeon/{dungeon_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_index(self) -> dict[str, Any]:
        """
        Returns an index of links to other documents related to Mythic Keystone dungeons.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = "/data/wow/mythic-keystone/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_period_index(self) -> dict[str, Any]:
        """
        Returns an index of Mythic Keystone periods.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = "/data/wow/mythic-keystone/period/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_period(self, period_id: int) -> dict[str, Any]:
        """
        Returns a Mythic Keystone period by ID.

        :param period_id: The ID of the Mythic Keystone season period.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/mythic-keystone/period/{period_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_season_index(self) -> dict[str, Any]:
        """
        Returns an index of Mythic Keystone seasons.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = "/data/wow/mythic-keystone/season/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_season(self, season_id: int) -> dict[str, Any]:
        """
        Returns a Mythic Keystone season by ID.

        :param season_id: The ID of the Mythic Keystone season.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/mythic-keystone/season/{season_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Mythic Keystone Leaderboard API
    def mythic_keystone_leaderboard_index(self, connected_realm_id: int) -> dict[str, Any]:
        """
        Returns an index of Mythic Keystone Leaderboard dungeon instances for a connected realm.

        :param connected_realm_id: The ID of the connected realm.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def mythic_keystone_leaderboard(self, connected_realm_id: int, dungeon_id: int, period_id: int) -> dict[str, Any]:
        """
        Returns a weekly Mythic Keystone Leaderboard by period.

        :param connected_realm_id: The ID of the connected realm.
        :param dungeon_id: The ID of the dungeon.
        :param period_id: The unique identifier for the leaderboard period.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/connected-realm/{connected_realm_id}/mythic-leaderboard/" \
              f"{dungeon_id}/period/{period_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Mythic Raid Leaderboard API
    def mythic_raid_leaderboard(self, raid: str, faction: str) -> dict[str, Any]:
        """
        Returns the leaderboard for a given raid and faction.

        :param raid: The raid for a leaderboard.
        :param faction: Player faction (`alliance` or `horde`).
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/leaderboard/hall-of-fame/{raid}/{faction}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Pet API
    def pets_index(self) -> dict[str, Any]:
        """
        Returns an index of battle pets.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/pet/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pet(self, pet_id: int) -> dict[str, Any]:
        """
        Returns a battle pet by ID.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/pet/{pet_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pet_media(self, pet_id: int) -> dict[str, Any]:
        """
        Returns media for a battle pet by ID.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/pet/{pet_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pet_abilities_index(self) -> dict[str, Any]:
        """
        Returns an index of pet abilities.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/pet-ability/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pet_ability(self, pet_ability_id: int) -> dict[str, Any]:
        """
        Returns a pet ability by ID.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/pet-ability/{pet_ability_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pet_ability_media(self, pet_ability_id: int) -> dict[str, Any]:
        """
        Returns media for a pet ability by ID.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/pet-ability/{pet_ability_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Playable Class APIs
    def playable_classes_index(self) -> dict[str, Any]:
        """
        Returns an index of playable classes.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/playable-class/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def playable_class(self, playable_class_id: int) -> dict[str, Any]:
        """
        Returns a playable class by ID.

        :param playable_class_id: The ID of the playable class.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/playable-class/{playable_class_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def playable_class_media(self, playable_class_id: int) -> dict[str, Any]:
        """
        Returns media for a playable class by ID.

        :param playable_class_id: The ID of the playable class.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/media/playable-class/{playable_class_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_talent_slots(self, playable_class_id: int) -> dict[str, Any]:
        """
        Returns the PvP talent slots for a playable class by ID.

        :param playable_class_id: The ID of the playable class.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/playable-class/{playable_class_id}/pvp-talent-slots"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Playable Race APIs
    def playable_races_index(self) -> dict[str, Any]:
        """
        Returns an index of playable races.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/playable-race/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def playable_race(self, playable_race_id: int) -> dict[str, Any]:
        """
        Returns a playable race by ID.

        :param playable_race_id: The ID of the playable race.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/playable-race/{playable_race_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Playable Specialization APIs
    def playable_specializations_index(self) -> dict[str, Any]:
        """
        Returns an index of playable specializations.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/playable-specialization/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def playable_specialization(self, playable_specialization_id: int) -> dict[str, Any]:
        """
        Returns a playable specialization by ID.

        :param playable_specialization_id: The ID of the playable specialization.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/playable-specialization/{playable_specialization_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def playable_specialization_media(self, playable_specialization_id: int) -> dict[str, Any]:
        """
        Returns media for a playable specialization by ID.

        :param playable_specialization_id: The ID of the playable specialization.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/playable-specialization/{playable_specialization_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Power Type APIs
    def power_types_index(self) -> dict[str, Any]:
        """
        Returns an index of power types.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = "/data/wow/power-type/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def power_type(self, power_type_id: int) -> dict[str, Any]:
        """
        Returns a power type by ID.

        :param power_type_id: The ID of the power type.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.STATIC
        url = f"/data/wow/power-type/{power_type_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Profession APIs
    def professions_index(self) -> dict[str, Any]:
        """
        Returns an index of professions.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/profession/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def profession(self, profession_id: int) -> dict[str, Any]:
        """
        Returns a profession by ID.

        :param profession_id: The ID of the profession.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/profession/{profession_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def profession_media(self, profession_id: int) -> dict[str, Any]:
        """
        Returns a profession by ID.

        :param profession_id: The ID of the profession.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/profession/{profession_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def profession_skill_tier(self, profession_id: int, skill_tier_id: int) -> dict[str, Any]:
        """
        Returns a skill tier for a profession by ID.

        :param profession_id: The ID of the profession.
        :param skill_tier_id: The ID of the skill tier.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/profession/{profession_id}/skill-tier/{skill_tier_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def recipe(self, recipe_id: int) -> dict[str, Any]:
        """
        Returns a recipe by ID.

        :param recipe_id: The ID of the recipe.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/recipe/{recipe_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def recipe_media(self, recipe_id: int) -> dict[str, Any]:
        """
        Returns media for a recipe by ID.

        :param recipe_id: The ID of the recipe.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/recipe/{recipe_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # PvP Season APIs
    def pvp_seasons_index(self) -> dict[str, Any]:
        """
        Returns an index of PvP seasons.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC
        namespace = Namespace.DYNAMIC
        url = "/data/wow/pvp-season/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_season(self, pvp_season_id: int) -> dict[str, Any]:
        """
        Returns a PvP season by ID.

        :param pvp_season_id: The ID of the PvP season.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/pvp-season/{pvp_season_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_leaderboards_index(self, pvp_season_id: int) -> dict[str, Any]:
        """
        Returns an index of PvP leaderboards for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/pvp-season/{pvp_season_id}/pvp-leaderboard/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_leaderboard(self, pvp_season_id: int, pvp_bracket: str) -> dict[str, Any]:
        """
        Returns the PvP leaderboard of a specific PvP bracket for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        :param pvp_bracket: The PvP bracket type.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/pvp-season/{pvp_season_id}/pvp-leaderboard/{pvp_bracket}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_rewards_index(self, pvp_season_id: int) -> dict[str, Any]:
        """
        Returns an index of PvP rewards for a PvP season.

        :param pvp_season_id: The ID of the PvP season.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/pvp-season/{pvp_season_id}/pvp-reward/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # PvP Tier APIs
    def pvp_tier_media(self, pvp_tier_id: int) -> dict[str, Any]:
        """
        Returns media for a PvP tier by ID.

        :param pvp_tier_id: The ID of the PvP tier.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/pvp-tier/{pvp_tier_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_tiers_index(self) -> dict[str, Any]:
        """
        Returns an index of PvP tiers.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/pvp-tier/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_tier(self, pvp_tier_id: int) -> dict[str, Any]:
        """
        Returns a PvP tier by ID.

        :param pvp_tier_id: The ID of the PvP tier.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/pvp-tier/{pvp_tier_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Quest APIs
    def quests_index(self) -> dict[str, Any]:
        """
        Returns the parent index for quests.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/quest/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest(self, quest_id: int) -> dict[str, Any]:
        """
        Returns a quest by ID.

        :param quest_id: The ID of the quest.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/quest/{quest_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_categories_index(self) -> dict[str, Any]:
        """
        Returns an index of quest categories (such as quests for a specific class, profession, or storyline).
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/quest/category/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_category(self, quest_category_id: int) -> dict[str, Any]:
        """
        Returns a quest category by ID.

        :param quest_category_id: The ID of the quest category.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/quest/category/{quest_category_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_areas_index(self) -> dict[str, Any]:
        """
        Returns an index of quest areas.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/quest/area/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_area(self, quest_area_id: int) -> dict[str, Any]:
        """
        Returns a quest area by ID.

        :param quest_area_id: The ID of the quest area.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/quest/area/{quest_area_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_types_index(self) -> dict[str, Any]:
        """
        Returns an index of quest types (such as PvP quests, raid quests, or account quests).
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/quest/type/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def quest_type(self, quest_type_id: int) -> dict[str, Any]:
        """
        Returns a quest type by ID.

        :param quest_type_id: The ID of the quest type.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/quest/type/{quest_type_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Realm APIs
    def realms_index(self) -> dict[str, Any]:
        """
        Returns an index of realms.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = "/data/wow/realm/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def realm(self, realm_id: int | str) -> dict[str, Any]:
        """
        Returns a single realm by slug or ID.

        :param realm_id: The slug or ID of the realm.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/realm/{realm_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def realm_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of realms.
        """
        raise NotImplementedError()

    # Region APIs
    def regions_index(self) -> dict[str, Any]:
        """
        Returns an index of regions.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = "/data/wow/region/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def region(self, region_id: int) -> dict[str, Any]:
        """
        Returns a region by ID.

        :param region_id: The ID of the region.
        """
        expansions = WarcraftExpansion.RETAIL, WarcraftExpansion.TBCC, WarcraftExpansion.CLASSIC
        namespace = Namespace.DYNAMIC
        url = f"/data/wow/region/{region_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Reputations APIs
    def reputation_factions_index(self) -> dict[str, Any]:
        """
        Returns an index of reputation factions.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/reputation-faction/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def reputation_faction(self, reputation_faction_id: int) -> dict[str, Any]:
        """
        Returns a single reputation faction by ID.

        :param reputation_faction_id: The ID of the reputation faction.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/reputation-faction/{reputation_faction_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def reputation_tiers_index(self) -> dict[str, Any]:
        """
        Returns an index of reputation tiers.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/reputation-tiers/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def reputation_tier(self, reputation_tier_id: int) -> dict[str, Any]:
        """
        Returns a single set of reputation tiers by ID.

        :param reputation_tier_id: The ID of the reputation faction.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/reputation-tiers/{reputation_tier_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Spell API
    def spell(self, spell_id: int) -> dict[str, Any]:
        """
        Returns a spell by ID.

        :param spell_id: The ID of the spell.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/spell/{spell_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def spell_media(self, spell_id: int) -> dict[str, Any]:
        """
        Returns media for a spell by ID.

        :param spell_id: The ID of the spell.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/spell/{spell_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def spell_search(self, **kwargs: dict) -> None:
        """
        NOT IMPLEMENTED. Performs a search of spells.
        """
        raise NotImplementedError()

    # Talent APIs
    def talents_index(self) -> dict[str, Any]:
        """
        Returns an index of talents.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/talent/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def talent(self, talent_id: int) -> dict[str, Any]:
        """
        Returns a talent by ID.

        :param talent_id: The ID of the talent.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/talent/{talent_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_talents_index(self) -> dict[str, Any]:
        """
        Returns an index of PvP talents.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/pvp-talent/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def pvp_talent(self, pvp_talent_id: int) -> dict[str, Any]:
        """
        Returns a PvP talent by ID.

        :param pvp_talent_id: The ID of the PvP talent.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/pvp-talent/{pvp_talent_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Tech Talent APIs
    def tech_talent_tree_index(self) -> dict[str, Any]:
        """
        Returns an index of tech talent trees.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/tech-talent-tree/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def tech_talent_tree(self, tech_talent_tree_id: int) -> dict[str, Any]:
        """
        Returns a tech talent tree by ID.

        :param tech_talent_tree_id: The ID of the tech talent tree.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/tech-talent-tree/{tech_talent_tree_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def tech_talent_index(self) -> dict[str, Any]:
        """
        Returns an index of tech talents.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/tech-talent/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def tech_talent(self, tech_talent_id: int) -> dict[str, Any]:
        """
        Returns a tech talent by ID.

        :param tech_talent_id: The ID of the tech talent.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/tech-talent/{tech_talent_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def tech_talent_media(self, tech_talent_id: int) -> dict[str, Any]:
        """
        Returns media for a tech talent by ID.

        :param tech_talent_id: The ID of the tech talent.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/media/tech-talent/{tech_talent_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Title APIs
    def titles_index(self) -> dict[str, Any]:
        """
        Returns an index of titles.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = "/data/wow/title/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def title(self, title_id: int) -> dict[str, Any]:
        """
        Returns a title by ID.

        :param title_id: The ID of the title.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.STATIC
        url = f"/data/wow/title/{title_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Token APIs
    def token(self) -> dict[str, Any]:
        """
        Returns the WoW Token index.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.DYNAMIC
        url = "/data/wow/token/index"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Account Profile APIs
    def account_profile_summary(self) -> dict[str, Any]:
        """
        Returns a profile summary for an account.

        Because this endpoint provides data about the current logged-in user's World of Warcraft account,
        it requires an access token with the wow.profile scope
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = "/profile/user/wow"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def protected_character_profile_summary(self, realm_id: int, character_id: int) -> dict[str, Any]:
        """
        Returns a protected profile summary for a character.

        Because this endpoint provides data about the current logged-in user's World of Warcraft account,
        it requires an access token with the wow.profile scope

        :param realm_id: The ID of the character's realm.
        :param character_id: The ID of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/user/wow/protected-character/{realm_id}-{character_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def account_collections_index(self) -> dict[str, Any]:
        """
        Returns an index of collection types for an account.

        Because this endpoint provides data about the current logged-in user's World of Warcraft account,
        it requires an access token with the wow.profile scope
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = "/profile/user/wow/collections"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def account_mounts_collection_summary(self) -> dict[str, Any]:
        """
        Returns a summary of the mounts an account has obtained.

        Because this endpoint provides data about the current logged-in user's World of Warcraft account,
        it requires an access token with the wow.profile scope
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = "/profile/user/wow/collections/mounts"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def account_pets_collection_summary(self) -> dict[str, Any]:
        """
        Returns a summary of the battle pets an account has obtained.

        Because this endpoint provides data about the current logged-in user's World of Warcraft account,
        it requires an access token with the wow.profile scope
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = "/profile/user/wow/collections/pets"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Achievements API
    def character_achievements_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of the achievements a character has completed.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/achievements"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_achievement_statistics(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a character's statistics as they pertain to achievements.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/achievements/statistics"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Appearance API
    def character_appearance_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's appearance settings.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/appearance"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Collections APIs
    def character_collections_index(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns an index of collection types for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/collections"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_mounts_collection_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of the mounts a character has obtained.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/collections/mounts"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_pets_collection_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of the battle pets a character has obtained.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/collections/pets"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Encounters APIs
    def character_encounters_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's encounters.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/encounters"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_dungeons(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's completed dungeons.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/encounters/dungeons"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_raids(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's completed raids.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/encounters/raids"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Encounters APIs
    def character_equipment_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of the items equipped by a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/equipment"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Hunter Pets APIs
    def character_hunter_pets_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        If the character is a hunter, returns a summary of the character's hunter pets.
        Otherwise, returns an HTTP 404 Not Found error.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/hunter-pets"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Media APIs
    def character_media_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of the media assets available for a character (such as an avatar render).

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/character-media"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Mythic Keystone Profile APIs
    def character_mythic_keystone_profile_index(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns the Mythic Keystone profile index for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/mythic-keystone-profile"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_mythic_keystone_season_details(
            self,
            realm_slug: str,
            character_name: str,
            season_id: int
    ) -> dict[str, Any]:
        """
        Returns the Mythic Keystone season details for a character.

        Returns a 404 Not Found for characters that have not yet completed
        a Mythic Keystone dungeon for the specified season.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        :param season_id: The ID of the Mythic Keystone season.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/mythic-keystone-profile/season/{season_id}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Professions APIs
    def character_professions_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of professions for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/professions"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Profile APIs
    def character_profile_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a profile summary for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_profile_status(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns the status and a unique ID for a character.

        A client should delete information about a character from their application if any of the following occur:
            - An HTTP 404 Not Found error is returned
            - The is_valid value is false
            - The returned character ID doesn't match the previously recorded value for the character

        The following example illustrates how to use this endpoint:
            - A client requests and stores information about a character,
              including its unique character ID and the timestamp of the request.
            - After 30 days, the client makes a request to the status endpoint to
              verify if the character information is still valid.
            - If character cannot be found, is not valid, or the characters IDs do not match,
              the client removes the information from their application.
            - If the character is valid and the character IDs match, the client retains the data for another 30 days.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/status"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character PvP APIs
    def character_pvp_bracket_statistics(
            self,
            realm_slug: str,
            character_name: str,
            pvp_bracket: str
    ) -> dict[str, Any]:
        """
        Returns a profile summary for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        :param pvp_bracket: The PvP bracket type.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/pvp-bracket/{pvp_bracket}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_pvp_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a PvP summary for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/pvp-summary"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Quest APIs
    def character_quests(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a character's active quests as well as a link to the character's completed quests.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/quests"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def character_completed_quests(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a list of quests that a character has completed.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/quests/completed"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Reputations APIs
    def character_reputations_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's reputations.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/reputations"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Soulbinds APIs
    def character_soulbinds(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a character's soulbinds.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/soulbinds"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Specializations APIs
    def character_specializations_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of a character's specializations.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/specializations"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Statistics APIs
    def character_statistics_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a statistics summary for a character.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/statistics"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Character Titles APIs
    def character_titles_summary(self, realm_slug: str, character_name: str) -> dict[str, Any]:
        """
        Returns a summary of titles a character has obtained.

        :param realm_slug: The slug of the realm.
        :param character_name: The lowercase name of the character.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/profile/wow/character/{realm_slug}/{character_name}/titles"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    # Guild APIs
    def guild(self, realm_slug: str, guild_slug: str) -> dict[str, Any]:
        """
        Returns a summary of titles a character has obtained.

        :param realm_slug: The slug of the realm.
        :param guild_slug: The slug of the guild.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/data/wow/guild/{realm_slug}/{guild_slug}"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def guild_activity(self, realm_slug: str, guild_slug: str) -> dict[str, Any]:
        """
        Returns a single guild's activity by name and realm.

        :param realm_slug: The slug of the realm.
        :param guild_slug: The slug of the guild.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/data/wow/guild/{realm_slug}/{guild_slug}/activity"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def guild_achievements(self, realm_slug: str, guild_slug: str) -> dict[str, Any]:
        """
        Returns a single guild's achievements by name and realm.

        :param realm_slug: The slug of the realm.
        :param guild_slug: The slug of the guild.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/data/wow/guild/{realm_slug}/{guild_slug}/achievements"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)

    def guild_roster(self, realm_slug: str, guild_slug: str) -> dict[str, Any]:
        """
        Returns a single guild's roster by its name and realm.

        :param realm_slug: The slug of the realm.
        :param guild_slug: The slug of the guild.
        """
        expansions = WarcraftExpansion.RETAIL,
        namespace = Namespace.PROFILE
        url = f"/data/wow/guild/{realm_slug}/{guild_slug}/roster"

        return self.prepare(expansions=expansions, namespace=namespace, url=url)
